import os
import numpy as np
import re
import matplotlib.pyplot as plt
import matplotlib.pylab as pl
from matplotlib.animation import FuncAnimation
import pandas as pd
from scipy import integrate, interpolate
from typing import (List, Literal, Callable)
from pydantic import BaseModel

from fiqus.utils.Utils import FilesAndFolders as Util
from fiqus.data.DataFiQuSConductorAC_Strand import CACStrandSolve, CACStrandPostproc, CACStrandMesh, CACStrandGeometry
from fiqus.plotters.PlotPythonConductorAC import PlotPython, SimulationData


class PostProcess:
    """
    This class loads and stores the data from a simulation and can apply various postprocessing operations on the data. 
    The simulation data is saved as a SimulationData object.
    """
    def __init__(self, fdm, model_data_output_path) -> None:
        self.fdm = fdm
        self.model_data_output_path = model_data_output_path
        self.geometry_name = f"Geometry_{self.fdm.run.geometry}"
        self.mesh_name = f"Mesh_{self.fdm.run.mesh}"
        self.solution_name = f"Solution_{self.fdm.run.solution}"

        self.simulation_data = SimulationData(model_data_output_path, self.geometry_name, self.mesh_name, self.solution_name)

    def plot_instantaneous_loss(self):
        print("Total loss: ", self.simulation_data.cumulative_power["TotalLoss"].iloc[-1])
        print("Total filament loss: ", self.simulation_data.cumulative_power["FilamentLoss"].iloc[-1])
        print("Total coupling loss: ", self.simulation_data.cumulative_power["CouplingLoss"].iloc[-1])
        print("Total eddy loss: ", self.simulation_data.cumulative_power["EddyLoss"].iloc[-1])
        plot_options = self.fdm.magnet.postproc.plot_instantaneous_power
        self.simulation_data.plot_instantaneous_power(
            show=plot_options.show, 
            title=plot_options.title, 
            save_plot=plot_options.save, 
            save_folder_path=os.path.join(self.model_data_output_path, self.geometry_name, self.mesh_name, self.solution_name),
            save_file_name=plot_options.save_file_name,
            overwrite=self.fdm.run.overwrite
            )








import os
import gmsh
import time

from fiqus.utils.Utils import GmshUtils
from fiqus.utils.Utils import FilesAndFolders as Util
from fiqus.data import DataFiQuS as dF
from fiqus.geom_generators.GeometryMultipole import Geometry
from fiqus.mesh_generators.MeshMultipole import Mesh
from fiqus.getdp_runners.RunGetdpMultipole import RunGetdpMultipole
from fiqus.getdp_runners.RunGetdpMultipole import AssignNaming
from fiqus.post_processors.PostProcessMultipole import PostProcess
from fiqus.plotters.PlotPythonMultipole import PlotPythonMultipole


class MainMultipole:
    def __init__(self, fdm: dF.FDM = None, rgd_path: str = None, verbose: bool = None):
        """
        Main class for working with simulations for multipole type magnets
        :param fdm: FiQuS data model
        :param rgd_path: ROXIE geometry data path
        :param verbose: if True, more info is printed in the console
        """
        self.fdm = fdm
        self.rgd = rgd_path
        self.verbose = verbose

        self.GetDP_path = None
        self.geom_folder = None
        self.mesh_folder = None
        self.solution_folder = None

    def force_symmetry(self):
        fdm = self.fdm.__deepcopy__()
        fdm.magnet.geometry.electromagnetics.symmetry = 'x'
        return fdm
    def generate_geometry(self, gui: bool = False):
        geom = Util.read_data_from_yaml(self.rgd, dF.FiQuSGeometry)
        fdm = self.force_symmetry() if 'solenoid' in geom.Roxie_Data.coil.coils[1].type else self.fdm  # todo: this should be handled by pydantic
        if self.fdm.magnet.geometry.plot_preview:
            plotter = PlotPythonMultipole(geom, self.fdm)
            plotter.plot_coil_wedges()
        gg = Geometry(data=fdm, geom=geom, geom_folder=self.geom_folder, verbose=self.verbose)
        gg.saveHalfTurnCornerPositions()
        geometry_settings = {'EM': fdm.magnet.geometry.electromagnetics, 'TH': self.fdm.magnet.geometry.thermal}
        geometry_type_list = []
        if fdm.magnet.geometry.electromagnetics.create: geometry_type_list.append('EM')
        if fdm.magnet.geometry.thermal.create: geometry_type_list.append('TH')
        for geometry_type in geometry_type_list:
            gg.saveStrandPositions(geometry_type)
            if geometry_settings[geometry_type].with_iron_yoke:
                gg.constructIronGeometry(geometry_settings[geometry_type].symmetry if geometry_type == 'EM' else 'none')
            gg.constructCoilGeometry(geometry_type)
            if geometry_settings[geometry_type].with_wedges:
                gg.constructWedgeGeometry(geometry_settings[geometry_type].use_TSA if geometry_type == 'TH' else False)
            gmsh.model.occ.synchronize()
            if geometry_type == 'TH':
                if geometry_settings[geometry_type].use_TSA:
                    gg.constructThinShells(geometry_settings[geometry_type].with_wedges)
                else:
                    gg.constructInsulationGeometry()
            gg.buildDomains(geometry_type, geometry_settings[geometry_type].symmetry if geometry_type == 'EM' else 'none')
            if geometry_type == 'EM':
                gg.fragment()
            gg.saveBoundaryRepresentationFile(geometry_type)
            gg.loadBoundaryRepresentationFile(geometry_type)
            gg.updateTags(geometry_type, geometry_settings[geometry_type].symmetry if geometry_type == 'EM' else 'none')
            gg.saveAuxiliaryFile(geometry_type)
            gg.clear()
        gg.ending_step(gui)

    def load_geometry(self, gui: bool = False):
        pass
        # gu = GmshUtils(self.geom_folder, self.verbose)
        # gu.initialize(verbosity_Gmsh=self.fdm.run.verbosity_Gmsh)
        # model_file = os.path.join(self.geom_folder, self.fdm.general.magnet_name)
        # gmsh.option.setString(name='Geometry.OCCTargetUnit', value='M')  # set units to meters
        # gmsh.open(model_file + '_EM.brep')
        # gmsh.open(model_file + '_TH.brep')
        # if gui: gu.launch_interactive_GUI()

    def pre_process(self, gui: bool = False):
        pass

    def load_geometry_for_mesh(self, run_type):
        gu = GmshUtils(self.geom_folder, self.verbose)
        gu.initialize(verbosity_Gmsh=self.fdm.run.verbosity_Gmsh)
        model_file = os.path.join(self.geom_folder, self.fdm.general.magnet_name)
        gmsh.option.setString(name='Geometry.OCCTargetUnit', value='M')  # set units to meters
        gmsh.open(model_file + f'_{run_type}.brep')

    def mesh(self, gui: bool = False):
        mm = Mesh(data=self.fdm, mesh_folder=self.mesh_folder, verbose=self.verbose)
        geom = Util.read_data_from_yaml(self.rgd, dF.FiQuSGeometry)
        fdm = self.force_symmetry() if 'solenoid' in geom.Roxie_Data.coil.coils[1].type else self.fdm
        geometry_settings = {'EM': fdm.magnet.geometry.electromagnetics, 'TH': self.fdm.magnet.geometry.thermal}
        mesh_settings = {'EM': fdm.magnet.mesh.electromagnetics, 'TH': fdm.magnet.mesh.thermal}
        mesh_type_list = []
        if fdm.magnet.mesh.electromagnetics.create: mesh_type_list.append('EM')
        if fdm.magnet.mesh.thermal.create: mesh_type_list.append('TH')
        for physics_solved in mesh_type_list:
            self.load_geometry_for_mesh(physics_solved)
            if physics_solved == 'TH' and self.fdm.magnet.geometry.thermal.use_TSA:
                mm.loadStrandPositions(physics_solved)
            mm.loadAuxiliaryFile(physics_solved)
            if geometry_settings[physics_solved].with_iron_yoke:
                mm.getIronCurvesTags()
            mm.defineMesh(geometry_settings[physics_solved], mesh_settings[physics_solved], physics_solved)
            mm.createPhysicalGroups(geometry_settings[physics_solved])
            mm.updateAuxiliaryFile(physics_solved)
            if geometry_settings[physics_solved].dict().get('use_TSA', False):
                mm.rearrangeThinShellsData()
            mm.assignRegionsTags(geometry_settings[physics_solved], mesh_settings[physics_solved])
            mm.saveRegionFile(physics_solved)
            mm.setMeshOptions(physics_solved)
            mm.generateMesh()
            mm.checkMeshQuality()
            mm.saveMeshFile(physics_solved)
            if geometry_settings[physics_solved].dict().get('use_TSA', False):
                mm.saveClosestNeighboursList()
                if self.fdm.magnet.mesh.thermal.isothermal_conductors: mm.selectMeshNodes(elements='conductors')
                if self.fdm.magnet.geometry.thermal.with_wedges and self.fdm.magnet.mesh.thermal.isothermal_wedges: mm.selectMeshNodes(elements='wedges')
                mm.saveRegionCoordinateFile(physics_solved)
            mm.clear()
        mm.ending_step(gui)
        return mm.mesh_parameters

    def load_mesh(self, gui: bool = False):
        gu = GmshUtils(self.geom_folder, self.verbose)
        gu.initialize(verbosity_Gmsh=self.fdm.run.verbosity_Gmsh)
        gmsh.open(f"{os.path.join(self.mesh_folder, self.fdm.general.magnet_name)}.msh")
        if gui: gu.launch_interactive_GUI()

    def solve_and_postprocess_getdp(self, gui: bool = False):
        an = AssignNaming(data=self.fdm)
        rg = RunGetdpMultipole(data=an, solution_folder=self.solution_folder, GetDP_path=self.GetDP_path, verbose=self.verbose)
        rg.loadRegionFiles()
        if self.fdm.magnet.solve.thermal.solve_type and self.fdm.magnet.geometry.thermal.use_TSA:
            rg.loadRegionCoordinateFile()
        rg.assemblePro()
        start_time = time.time()
        rg.solve_and_postprocess()
        rg.ending_step(gui)
        return time.time() - start_time

    def post_process_getdp(self, gui: bool = False):
        an = AssignNaming(data=self.fdm)
        rg = RunGetdpMultipole(data=an, solution_folder=self.solution_folder, GetDP_path=self.GetDP_path, verbose=self.verbose)
        rg.loadRegionFiles()
        if self.fdm.magnet.solve.thermal.solve_type and self.fdm.magnet.geometry.thermal.use_TSA:
            rg.loadRegionCoordinateFile()
        rg.assemblePro()
        rg.postprocess()
        rg.ending_step(gui)

    def post_process_python(self, gui: bool = False):
        if self.fdm.run.type == 'post_process_python_only':
            an = AssignNaming(data=self.fdm)
            data = an.data
        else: data = self.fdm

        run_types = []
        if self.fdm.magnet.solve.electromagnetics.solve_type: run_types.append('EM')
        if self.fdm.magnet.solve.thermal.solve_type: run_types.append('TH')
        pp_settings = {'EM': self.fdm.magnet.postproc.electromagnetics, 'TH': self.fdm.magnet.postproc.thermal}
        pp = PostProcess(data=data, solution_folder=self.solution_folder, verbose=self.verbose)
        for run_type in run_types:
            pp.prepare_settings(pp_settings[run_type])
            pp.loadStrandPositions(run_type)
            pp.loadAuxiliaryFile(run_type)
            if pp_settings[run_type].plot_all != 'False': pp.loadHalfTurnCornerPositions()
            if pp_settings[run_type].dict().get('take_average_conductor_temperature', False): pp.loadRegionFile()
            pp.postProcess(pp_settings[run_type])
            if run_type == 'EM' and self.fdm.magnet.geometry.electromagnetics.symmetry != 'none': pp.completeMap2d()
            pp.clear()
        pp.ending_step(gui)
        return pp.postprocess_parameters

    def plot_python(self):
        os.chdir(self.solution_folder)
        p = PlotPythonMultipole(self.fdm, self.fdm)
        p.plot_coil_wedges()



import timeit
import logging
import os
import subprocess
import re
import pandas as pd
import pickle
import numpy as np

import gmsh

from fiqus.utils.Utils import GmshUtils, FilesAndFolders
from fiqus.data.RegionsModelFiQuS import RegionsModel
import fiqus.data.DataFiQuSConductor as geom
from fiqus.geom_generators.GeometryConductorAC_Strand import TwistedStrand

from fiqus.pro_assemblers.ProAssembler import ASS_PRO

logger = logging.getLogger(__name__)

class Solve:
    def __init__(self, fdm, GetDP_path, geometry_folder, mesh_folder, verbose=True):
        self.fdm = fdm
        self.cacdm = fdm.magnet
        self.GetDP_path = GetDP_path
        self.solution_folder = os.path.join(os.getcwd())
        self.magnet_name = fdm.general.magnet_name
        self.geometry_folder = geometry_folder
        self.mesh_folder = mesh_folder
        self.mesh_file = os.path.join(self.mesh_folder, f"{self.magnet_name}.msh")
        self.pro_file = os.path.join(self.solution_folder, f"{self.magnet_name}.pro")
        self.regions_file = os.path.join(mesh_folder, f"{self.magnet_name}.regions")
        
        self.verbose = verbose
        self.gu = GmshUtils(self.solution_folder, self.verbose)
        self.gu.initialize()

        self.ass_pro = ASS_PRO(os.path.join(self.solution_folder, self.magnet_name))
        self.regions_model = FilesAndFolders.read_data_from_yaml(self.regions_file, RegionsModel)
        self.material_properties_model = None

        self.ed = {} # excitation dictionary

        gmsh.option.setNumber("General.Terminal", verbose)

    def read_excitation(self, inputs_folder_path):
        """
        Function for reading a CSV file for the 'from_file' excitation case.

        :param inputs_folder_path: The full path to the folder with input files.
        :type inputs_folder_path: str
        """
        if self.cacdm.solve.source_parameters.source_type == 'piecewise' and self.cacdm.solve.source_parameters.piecewise.source_csv_file:
            input_file = os.path.join(inputs_folder_path, self.cacdm.solve.source_parameters.piecewise.source_csv_file)
            print(f'Using excitation from file: {input_file}')
            df = pd.read_csv(input_file, delimiter=',', engine='python')
            excitation_time = df['time'].to_numpy(dtype='float').tolist()
            self.ed['time'] = excitation_time
            excitation_value = df['value'].to_numpy(dtype='float').tolist()
            self.ed['value'] = excitation_value

    def get_solution_parameters_from_yaml(self, inputs_folder_path):
        """
        Function for reading material properties from the geometry YAML file.

        This reads the 'solution' section of the YAML file and stores it in the solution folder.
        This could also be a place to change the material properties in the future.

        :param inputs_folder_path: The full path to the folder with input files.
        :type inputs_folder_path: str
        """
        if self.cacdm.geometry.io_settings.load.load_from_yaml:
            #load the geometry class from the pkl file
            geom_save_file = os.path.join(self.geometry_folder, f'{self.magnet_name}.pkl')
            with open(geom_save_file, "rb") as geom_save_file: 
                geometry_class: TwistedStrand = pickle.load(geom_save_file)
            
            input_yaml_file = os.path.join(inputs_folder_path, self.cacdm.geometry.io_settings.load.filename)
            Conductor_dm = FilesAndFolders.read_data_from_yaml(input_yaml_file, geom.Conductor)
            solution_parameters = Conductor_dm.Solution

            # The geometry YAML file lists surfaces to exclude from the TI problem by their IDs. Here we convert these IDs to Gmsh physical surface tags.
            # This is done by comparing the outer boundary points of the surfaces to exclude with the outer boundary points of the matrix partitions.
            surfaces_excluded_from_TI_tags = []

            for surface_ID in solution_parameters.Surfaces_excluded_from_TI: # 1) Find the outer boundary points of the surfaces to exclude
                outer_boundary_points_a = []
                outer_boundary_curves_a = Conductor_dm.Geometry.Areas[surface_ID].Boundary
                for curve_ID in outer_boundary_curves_a:
                    curve = Conductor_dm.Geometry.Curves[curve_ID]
                    for point_ID in curve.Points:
                        point = Conductor_dm.Geometry.Points[point_ID]
                        outer_boundary_points_a.append(tuple(point.Coordinates))
                
                for matrix_partition in geometry_class.matrix: # 2) Find the outer boundary points of the matrix partitions
                    outer_boundary_points_b = []
                    outer_boundary_curves_b = matrix_partition.boundary_curves
                    if len(outer_boundary_curves_b) == len(outer_boundary_curves_a): # If the number of boundary curves is different, the surfaces are not the same
                        for curve in outer_boundary_curves_b:
                            for point in curve.points:
                                outer_boundary_points_b.append(tuple(point.pos))
                        
                        if np.allclose(sorted(outer_boundary_points_a), sorted(outer_boundary_points_b)): # If the outer boundary points are the same, the surfaces are the same
                            surfaces_excluded_from_TI_tags.append(matrix_partition.physical_surface_tag) # 3) Add the physical surface tag to the list of surfaces to exclude
                            break
            
            solution_parameters.Surfaces_excluded_from_TI = surfaces_excluded_from_TI_tags # Replace the surface IDs with the physical surface tags

            FilesAndFolders.write_data_to_yaml(os.path.join(self.solution_folder, "MaterialProperties.yaml"), solution_parameters.dict())
            self.material_properties_model = solution_parameters

        


    def assemble_pro(self):
        print("Assembling .pro file")
        self.ass_pro.assemble_combined_pro(template = self.cacdm.solve.pro_template, rm = self.regions_model, dm = self.fdm, ed=self.ed, mp=self.material_properties_model)

    def run_getdp(self, solve = True, postOperation = True, gui = False):

        command = ["-v2", "-verbose", "3"]
        if solve: 
            command += ["-solve", "MagDyn"]
        if self.cacdm.solve.formulation_parameters.dynamic_correction:
            command += ["-pos", "MagDyn", "MagDyn_dynCorr"]
        else:
            command += ["-pos", "MagDyn"]


        startTime = timeit.default_timer()
        getdpProcess = subprocess.Popen([self.GetDP_path, self.pro_file, "-msh", self.mesh_file] + command, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)

        with getdpProcess.stdout:
            for line in iter(getdpProcess.stdout.readline, b""):
                line = line.decode("utf-8").rstrip()
                line = line.split("\r")[-1]
                if not "Test" in line:
                    if line.startswith("Info"):
                        parsedLine = re.sub(r"Info\s+:\s+", "", line)
                        logger.info(parsedLine)
                    elif line.startswith("Warning"):
                        parsedLine = re.sub(r"Warning\s+:\s+", "", line)
                        logger.warning(parsedLine)
                    elif line.startswith("Error"):
                        parsedLine = re.sub(r"Error\s+:\s+", "", line)
                        logger.error(parsedLine)
                        logger.error("Solving CAC failed.")
                        # raise Exception(parsedLine)
                    elif re.match("##", line):
                        logger.critical(line)
                    else:
                        logger.info(line)
        
        simulation_time = timeit.default_timer()-startTime
        # Save simulation time:
        if solve:
            logger.info(f"Solving CAC_1 has finished in {round(simulation_time, 3)} seconds.")
            with open(os.path.join(self.solution_folder, 'test_temporary', 'simulation_time.txt'), 'w') as file:
                file.write(str(simulation_time))


        if gui and ((postOperation and not solve) or (solve and postOperation and self.cacdm.postproc.generate_pos_files)):
            # gmsh.option.setNumber("Geometry.Volumes", 1)
            # gmsh.option.setNumber("Geometry.Surfaces", 1)
            # gmsh.option.setNumber("Geometry.Curves", 1)
            # gmsh.option.setNumber("Geometry.Points", 0)
            posFiles = [
                fileName
                for fileName in os.listdir(self.solution_folder)
                if fileName.endswith(".pos")
            ]
            for posFile in posFiles:
                gmsh.open(os.path.join(self.solution_folder, posFile))
            self.gu.launch_interactive_GUI()
        else:
            gmsh.clear()
            gmsh.finalize()

    def cleanup(self):
        """
            This funtion is used to remove .msh, .pre and .res files from the solution folder, as they may be large and not needed.
        """
        magnet_name = self.fdm.general.magnet_name
        cleanup = self.cacdm.postproc.cleanup

        if cleanup.remove_res_file:
            res_file_path = os.path.join(self.solution_folder, f"{magnet_name}.res")
            if os.path.exists(res_file_path):
                os.remove(res_file_path)
                logger.info(f"Removed {magnet_name}.res")
        
        if cleanup.remove_pre_file:
            pre_file_path = os.path.join(self.solution_folder, f"{magnet_name}.pre")
            if os.path.exists(pre_file_path):
                os.remove(pre_file_path)
                logger.info(f"Removed {magnet_name}.pre")

        if cleanup.remove_msh_file:
            msh_file_path = os.path.join(self.mesh_folder, f"{magnet_name}.msh")
            if os.path.exists(msh_file_path):
                os.remove(msh_file_path)
                logger.info(f"Removed {magnet_name}.msh")


import argparse
import csv
import os
import pathlib
import sys
import time
import getpass
import platform
import subprocess
import json

import pandas as pd

FiQuS_path = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
sys.path.insert(0, FiQuS_path)

from fiqus.utils.Utils import FilesAndFolders as Util
from fiqus.utils.Utils import CheckForExceptions as Check
from fiqus.utils.Utils import create_json_schema
from fiqus.utils.Utils import get_data_settings
from fiqus.utils.Utils import initialize_logger
from fiqus.data.DataFiQuS import FDM
from fiqus.data.DataSettings import DataSettings
from fiqus.mains.MainCCT import MainCCT
from fiqus.mains.MainMultipole import MainMultipole
from fiqus.mains.MainPancake3D import MainPancake3D
from fiqus.mains.MainConductorAC_Strand import MainConductorAC_Strand

class MainFiQuS:
    """
    This is the top level class of FiQuS.
    """

    def __init__(
            self,
            input_file_path: str = None,
            model_folder: str = None,
            GetDP_path=None,
            fdm=None,
            fds=None,
            htcondor_jobid=None
    ):
        """
        Main class for working with FiQuS simulations
        :param input_file_path: full path to input file yaml
        :type input_file_path: str
        :param model_folder: full path to the base output folder, called model folder
        :type model_folder: str
        :param GetDP_path: full path to GetDP executable
        :type GetDP_path: str
        :param fdm: FiQuS Data Model - object of fiqus DataFiQus
        :type fdm: object
        :param fds: FiQuS Data Settings - object of DataSettings
        :type fds: object
        """
        self.time_stamp = time.strftime("%Y-%m-%d-%H-%M-%S")

        self.start_folder = os.getcwd()
        self.wrk_folder = model_folder
        self.file_name = None

        # Load yaml input file
        if not fdm:
            self.fdm: FDM = Util.read_data_from_yaml(input_file_path, FDM)
            copyInputFile = (
                "copy"
                f" {input_file_path} {os.path.join(self.wrk_folder, 'logs', f'INPUT_FILE_{self.time_stamp}.FiQuS.yaml')}"
            )
            subprocess.run(copyInputFile, shell=True, stdout=subprocess.DEVNULL)
        else:
            self.fdm: FDM = fdm
        verbose = self.fdm.run.verbosity_FiQuS
        self.logger = initialize_logger(
            verbose=verbose, time_stamp=self.time_stamp, work_folder=self.wrk_folder
        )
        if verbose:
            Util.print_welcome_graphics()
        # Intialize logger

        # Create JSON schema
        create_json_schema(self.fdm)

        # Check for input errors
        Check.check_inputs(run=self.fdm.run)

        # Initialize Main object
        if self.fdm.magnet.type == "CCT_straight":
            self.main_magnet = MainCCT(fdm=self.fdm, verbose=verbose)
        elif self.fdm.magnet.type == "CWS":
            self.main_magnet = MainCWS(
                fdm=self.fdm,
                inputs_folder_path=pathlib.Path(input_file_path).parent,
                verbose=self.fdm.run.verbosity_FiQuS,
            )
        elif self.fdm.magnet.type == "Pancake3D":
            self.main_magnet = MainPancake3D(fdm=self.fdm, verbose=verbose)
        elif self.fdm.magnet.type == "CACStrand":
            self.main_magnet = MainConductorAC_Strand(fdm=self.fdm, inputs_folder_path=pathlib.Path(input_file_path).parent, outputs_folder_path=model_folder, verbose=verbose)
        elif self.fdm.magnet.type == "CACRutherford":
            self.main_magnet = MainConductorAC_Rutherford(fdm=self.fdm, inputs_folder_path=pathlib.Path(input_file_path).parent, verbose=verbose)
        elif self.fdm.magnet.type == "Racetrack":
            self.main_magnet = MainRacetrack(fdm=self.fdm, verbose=verbose)
        elif self.fdm.magnet.type == "Racetrack3D":
            self.main_magnet = MainRacetrack3D(fdm=self.fdm, inputs_folder_path=pathlib.Path(input_file_path).parent, verbose=verbose)
        elif self.fdm.magnet.type == "multipole":
            self.file_name = os.path.basename(input_file_path)[:-5]
            if not self.fdm.magnet.geometry.geom_file_path:
                self.fdm.magnet.geometry.geom_file_path = f"{input_file_path[:-5]}.geom"
            self.main_magnet = MainMultipole(
                fdm=self.fdm,
                rgd_path=self.fdm.magnet.geometry.geom_file_path,
                verbose=verbose,
            )

        else:
            raise ValueError(
                f"FiQuS does not support magnet type: {self.fdm.magnet.type}!"
            )

        # Load user paths for executables and additional files
        self.logger.info(f'{getpass.getuser()} is running on {platform.platform()}')
        if not fds:
            fds = get_data_settings(GetDP_path=GetDP_path)
        else:
            fds = get_data_settings(GetDP_path=GetDP_path, settings=fds)
        self.main_magnet.GetDP_path = fds.GetDP_path
        self.logger.info(f"{self.main_magnet.GetDP_path} is going to be used for FE solving.")
        
        # self.logger.info(gmsh.onelab.run(self.fdm.general.magnet_name, f"{self.main_magnet.settings['GetDP_path']} -info"))
        
        # update htcondor csv
        if htcondor_jobid:
            base_path_model_files = fds.base_path_model_files
            htcondor_csv_file = os.path.join(base_path_model_files, "htcondor_run_log.csv")

            self.change_htcondor_run_log(htcondor_csv_file, htcondor_jobid, "Running")

        # Save Model/Geometry/Mesh/Solution folder paths
        self.save_folders()

        # Build magnet
        self.summary = dict.fromkeys(
            [
                "SJ",
                "SICN",
                "SIGE",
                "Gamma",
                "nodes",
                "solution_time",
                "overall_error",
                "minimum_diff",
                "maximum_diff",
            ]
        )

        try:
            self.build_magnet()
        except Exception as e:
            # update htcondor csv
            if htcondor_jobid:
                self.change_htcondor_run_log(htcondor_csv_file, htcondor_jobid, "Failed")
            
            self.logger.error(f"Error: {e}")
            raise e
        else:
            # update htcondor csv
            if htcondor_jobid:
                self.change_htcondor_run_log(htcondor_csv_file, htcondor_jobid, "Finished")

    def save_folders(self):
        """
        Method to make or delete folders of FiQuS
        :return: Nothing, only does file and folder operation
        :rtype: None
        """
        def _check_and_generate_path(folder_type: str = None):
            if folder_type == "Geometry":
                folder = self.wrk_folder
            elif folder_type == "Mesh":
                folder = self.main_magnet.geom_folder
            elif folder_type == "Solution":
                folder = self.main_magnet.mesh_folder
            else:
                raise Exception("Incompatible type.")

            if getattr(self.fdm.run, folder_type.lower()) is None:
                # folder_key is not given, so it is computed
                folder_key = Util.compute_folder_key(
                    folder_type=folder_type,
                    folder=folder,
                    overwrite=self.fdm.run.overwrite,
                )
            else:
                # folder_key is given
                folder_key = getattr(self.fdm.run, folder_type.lower())

            required_folder = folder_type in required_folders
            if self.fdm.run.overwrite and folder_type == (
                    required_folders[0] if required_folders else None
            ):
                Check.check_overwrite_conditions(
                    folder_type=folder_type, folder=folder, folder_key=folder_key
                )
            return Util.get_folder_path(
                folder_type=folder_type,
                folder=folder,
                folder_key=folder_key,
                overwrite=self.fdm.run.overwrite,
                required_folder=required_folder,
            )

        if self.fdm.run.type == "start_from_yaml":
            required_folders = ["Geometry", "Mesh", "Solution"]
        elif self.fdm.run.type == "geometry_and_mesh":
            required_folders = ["Geometry", "Mesh"]
        elif self.fdm.run.type == "mesh_and_solve_with_post_process_python":
            required_folders = ["Mesh", "Solution"]
        elif self.fdm.run.type in ["solve_with_post_process_python", "solve_only"]:
            required_folders = ["Solution"]
        elif self.fdm.run.type == "geometry_only":
            required_folders = (
                []
                if self.fdm.run.geometry and not self.fdm.run.overwrite
                else ["Geometry"]
            )
        elif self.fdm.run.type == "mesh_only":
            required_folders = (
                [] if self.fdm.run.mesh and not self.fdm.run.overwrite else ["Mesh"]
            )
        else:  # post_process_getdp_only or post_process_python_only or plot_python
            required_folders = []

        fdm = self.main_magnet.fdm.magnet

        self.main_magnet.geom_folder = _check_and_generate_path(folder_type="Geometry")
        if not self.fdm.run.type in ["geometry_only"]:
            self.main_magnet.mesh_folder = _check_and_generate_path(folder_type="Mesh")
        if not (
                self.fdm.run.type == "geometry_only"
                or self.fdm.run.type == "mesh_only"
        ):
            self.main_magnet.solution_folder = _check_and_generate_path(
                folder_type="Solution"
            )

        if self.fdm.run.type in [
            "start_from_yaml",
            "geometry_and_mesh",
            "geometry_only",
        ]:
            Util.write_data_model_to_yaml(
                os.path.join(self.main_magnet.geom_folder, "geometry.yaml"),
                fdm.geometry,
                by_alias=True,
                with_comments=True,
            )
        if self.fdm.run.type in [
            "start_from_yaml",
            "geometry_and_mesh",
            "mesh_and_solve_with_post_process_python",
            "mesh_only",
        ]:
            Util.write_data_model_to_yaml(
                os.path.join(self.main_magnet.mesh_folder, "mesh.yaml"),
                fdm.mesh,
                by_alias=True,
                with_comments=True,
            )
        if self.fdm.run.type in [
            "start_from_yaml",
            "mesh_and_solve_with_post_process_python",
            "solve_with_post_process_python",
            "solve_only",
            "post_process",
            "plot_python"
        ]:
            Util.write_data_model_to_yaml(
                os.path.join(self.main_magnet.solution_folder, "solve.yaml"),
                fdm.solve,
                by_alias=True,
                with_comments=True,
            )
        if self.fdm.run.type in [
            "start_from_yaml",
            "mesh_and_solve_with_post_process_python",
            "solve_with_post_process_python",
            "post_process_python_only",
            "post_process_getdp_only",
            "post_process",
            "plot_python"
        ]:
            Util.write_data_model_to_yaml(
                os.path.join(self.main_magnet.solution_folder, "postproc.yaml"),
                fdm.postproc,
                by_alias=True,
                with_comments=True,
            )

        try:
            run_type = self.fdm.run.type
            comments = self.fdm.run.comments
            if self.main_magnet.geom_folder is not None:
                geo_folder = os.path.relpath(self.main_magnet.geom_folder)
                geo_folder = os.path.relpath(
                    geo_folder, os.path.join("tests", "_outputs")
                )
            else:
                geo_folder = "-"

            if self.main_magnet.mesh_folder is not None:
                mesh_folder = os.path.relpath(self.main_magnet.mesh_folder)
                mesh_folder = os.path.relpath(
                    mesh_folder, os.path.join("tests", "_outputs")
                )
            else:
                mesh_folder = "-"

            if self.main_magnet.solution_folder is not None:
                solution_folder = os.path.relpath(self.main_magnet.solution_folder)
                solution_folder = os.path.relpath(
                    solution_folder, os.path.join("tests", "_outputs")
                )
            else:
                solution_folder = "-"

            run_log_row = [
                self.time_stamp,
                run_type,
                comments,
                geo_folder,
                mesh_folder,
                solution_folder,
            ]
            self.add_to_run_log(
                os.path.join(self.wrk_folder, "run_log.csv"), run_log_row
            )
        except:
            self.logger.warning("Run log could not be completed.")

    def build_magnet(self):
        """
        Main method to build magnets, i.e. to run various fiqus run types and magnet types
        :return: none
        :rtype: none
        """
        if self.fdm.run.type == "start_from_yaml":
            self.main_magnet.generate_geometry()
            self.main_magnet.pre_process()
            self.main_magnet.load_geometry()
            for key, value in self.main_magnet.mesh().items():
                self.summary[key] = value
            self.summary["solution_time"] = self.main_magnet.solve_and_postprocess_getdp()
            for key, value in self.main_magnet.post_process_python(gui=self.main_magnet.fdm.run.launch_gui).items():
                self.summary[key] = value
        elif self.fdm.run.type == "pre_process_only":
            self.main_magnet.pre_process()
            for key, value in self.main_magnet.post_process_python(gui=self.main_magnet.fdm.run.launch_gui).items():
                self.summary[key] = value  # todo: DISABLE FOR ONE GROUP ONLY
        elif self.fdm.run.type == "geometry_only":
            self.main_magnet.generate_geometry(
               gui=(self.main_magnet.fdm.run.launch_gui if self.fdm.magnet.type != "CCT_straight" else False)
            )
            if self.fdm.magnet.type in ["CCT_straight", "CWS"]:
                self.main_magnet.pre_process(gui=self.main_magnet.fdm.run.launch_gui)
        elif self.fdm.run.type == "geometry_and_mesh":
            self.main_magnet.generate_geometry()
            self.main_magnet.pre_process()
            self.main_magnet.load_geometry()
            for key, value in self.main_magnet.mesh(gui=self.main_magnet.fdm.run.launch_gui).items():
                self.summary[key] = value
        elif self.fdm.run.type == "mesh_and_solve_with_post_process_python":
            self.main_magnet.load_geometry()
            for key, value in self.main_magnet.mesh().items():
               self.summary[key] = value
            self.summary["solution_time"] = self.main_magnet.solve_and_postprocess_getdp()
            for key, value in self.main_magnet.post_process_python(gui=self.main_magnet.fdm.run.launch_gui).items():
               self.summary[key] = value
        elif self.fdm.run.type == "mesh_only":
            self.main_magnet.load_geometry()
            for key, value in self.main_magnet.mesh(gui=self.main_magnet.fdm.run.launch_gui).items():
                self.summary[key] = value
        elif self.fdm.run.type == "solve_with_post_process_python":
            self.summary["solution_time"] = (
                self.main_magnet.solve_and_postprocess_getdp(gui=self.main_magnet.fdm.run.launch_gui)
            )
            for key, value in self.main_magnet.post_process_python(gui=self.main_magnet.fdm.run.launch_gui).items():
                self.summary[key] = value
        elif self.fdm.run.type == "solve_only":
            self.summary["solution_time"] = (
                self.main_magnet.solve_and_postprocess_getdp(gui=self.main_magnet.fdm.run.launch_gui)
            )
        elif self.fdm.run.type == "post_process_getdp_only":
            self.main_magnet.post_process_getdp(gui=self.main_magnet.fdm.run.launch_gui)
        elif self.fdm.run.type == "post_process_python_only":
            for key, value in self.main_magnet.post_process_python(gui=self.main_magnet.fdm.run.launch_gui).items():
                self.summary[key] = value
        elif self.fdm.run.type == "post_process":
            self.main_magnet.post_process_getdp(gui=self.main_magnet.fdm.run.launch_gui)
            for key, value in self.main_magnet.post_process_python(gui=self.main_magnet.fdm.run.launch_gui).items():
                self.summary[key] = value
        elif self.fdm.run.type == "plot_python":
            self.main_magnet.plot_python()

        elif self.fdm.run.type == "batch_post_process_python":
            self.main_magnet.batch_post_process_python()
        os.chdir(self.start_folder)

        if self.file_name: json.dump(self.summary, open(f"{os.path.join(self.wrk_folder, self.file_name)}.json", 'w'))
        # mesh_par = self.fdm.magnet.mesh
        # if self.summary['solution_time']:
        #     with open(r"C:\Users\avitrano\PycharmProjects\steam_sdk\tests\parsims\FiQuS_run\summary.dat", 'a') as f:
        #         content = (f"{mesh_par.mesh_coil.SizeMin}     {mesh_par.mesh_coil.SizeMax}      {mesh_par.mesh_iron.SizeMin}      {mesh_par.mesh_iron.SizeMax}       "
        #                    f"{self.summary['solution_time']}    {self.summary['overall_error']}       {self.summary['overall_error'] * 0.999 + self.summary['solution_time'] * 0.001}       "
        #                    f"{self.summary['SJ']}     {self.summary['SICN']}      {self.summary['SIGE']}      {self.summary['Gamma']}       "
        #                    f"{self.summary['nodes']}    {self.summary['minimum_diff']}    {self.summary['maximum_diff']}\n")
        #         f.writelines(content)

    @staticmethod
    def add_to_run_log(path_to_csv, run_log_row):
        # If file does not exist, write the header
        if not os.path.isfile(path_to_csv):
            header = [
                "Time Stamp",
                "Run Type",
                "Comments",
                "Geometry Directory",
                "Mesh Directory",
                "Solution Directory",
            ]
            with open(path_to_csv, "a", newline="") as csv_file:
                writer = csv.writer(csv_file)
                writer.writerow(header)

        # Open the CSV file in append mode
        with open(path_to_csv, "a+", newline="") as csv_file:
            writer = csv.writer(csv_file)
            writer.writerow(run_log_row)

    def change_htcondor_run_log(self, htcondor_csv_file, htcondor_jobid, new_status="None"):
        try: 
            df = pd.read_csv(htcondor_csv_file)
            df.loc[df['Job ID'] == htcondor_jobid, 'Status'] = str(new_status)
            self.logger.info(f"Changed status of JobID {htcondor_jobid} to {new_status} in {htcondor_csv_file}.")
            df.to_csv(htcondor_csv_file, index=False)
        except:
            self.logger.warning(f"Could not change status of JobID {htcondor_jobid} to {new_status} in {htcondor_csv_file}.")

if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        prog="FiQuS",
        description="Finite Elements Quench Simulator",
        epilog="steam-team@cern.ch",
    )
    parser.add_argument(
        dest="full_path_input",
        type=str,
        help="Full path to FiQuS input yaml file",
    )
    parser.add_argument(
        "--output", '-o', dest="output_path", type=str, help="Full path to FiQuS output folder"
    )
    parser.add_argument(
        "--getdp", '-g', dest="GetDP_path", type=str, help="Full path to GetDP executable"
    )

    parser.add_argument("--htcondor_jobid", '-j', type=int, default=0,
                        help="HTCondor job ID (optional)", required=False)

    parser.add_argument("--fiqus_data_model", '-m', type=str, 
                        help="Full path to FiQuS Data Model file (optional)", required=False)

    parser.add_argument("--fiqus_data_settings", '-s', type=str, 
                        help="Full path to FiQuS Data Settings file (optional)", required=False)

    args, unknown = parser.parse_known_args()

    # remove these options from sys.argv, otherwise they are passed onto Gmsh
    # in Gmsh.initialize()
    options_to_remove = ["-o", "-g", "-j", "-m", "-s"]
    # Loop through and remove each option and its value
    i = 0
    while i < len(sys.argv):
        if sys.argv[i] in options_to_remove:
            sys.argv.pop(i)  # Remove the option
            if i < len(sys.argv):
                sys.argv.pop(i)  # Remove the associated value
        else:
            i += 1

    if args.fiqus_data_model != None and args.fiqus_data_settings != None:
        # read fdm and fds from a file (HTCondor case)
        input_fdm = Util.read_data_from_yaml(args.fiqus_data_model, FDM)
        input_fds = Util.read_data_from_yaml(args.fiqus_data_settings, DataSettings)

        MainFiQuS(
            input_file_path=args.full_path_input,
            model_folder=args.output_path,
            fdm=input_fdm,
            fds=input_fds,
            htcondor_jobid=args.htcondor_jobid
        )
    else:
        # fdm and fds from input (STEAM SDK case)
        MainFiQuS(
            input_file_path=args.full_path_input,
            model_folder=args.output_path,
            GetDP_path=args.GetDP_path,
        )
    print("FiQuS run completed")
